/*
authsyn.c,  proof-of-concept for authenticating syn packets to OpenBSD's pf 
using the built-in passive os fingerprinter.


for references see:
	http://tripp.dynalias.org/p0fspoof.txt
	
	
  usage:
      authsyn -i <interface> -h <host> -s <p0f signature> [-f <pcapfilter>] [-v]

      -i      output interface
      -h      destination host
      -s      signature from p0f databse
      -f      pcap filter string
      -v      be verbose

    example: 
     authsyn -i eth0 -h 85.41.32.67 -s "16384:64:1:60:M*,S,T,N,W0:" -f "port 80"

  note: if * is used, the original value is preserved
        it's possible to force T value by specifing a pair of values (tsval and
         echo-reply) separated by a dot, ex T1139773350.0
      

by poplix  poplix@papuasia.org   2006
      
*/



#define _BSD_SOURCE 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

//param.h defines BSD in bsd systems 
#include <sys/param.h>
#include <libnet.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef BSD
 #include <net/bpf.h>
#endif

#include <pcap.h>

#include <net/if.h>

#include <netinet/in_systm.h>
#include <netinet/tcp.h>
#include <netinet/ip.h>

#include <limits.h>
#include <errno.h>


#define VERSION "0.1"



#define BALIGN 0

#define MAXL2HDRSIZE 14

#define DEFAULT_SNAPLEN 96
#define MAXBUFFSIZE DEFAULT_SNAPLEN
#define OPTBUFF_SIZE 60

//rfc defined
#define TCPOPT_EOOL 	0
#define TCPOPT_NOOP		1
#define TCPOPT_MSS		2
#define TCPOPT_WSS		3
#define TCPOPT_SACKP	4
#define TCPOPT_TS		8



#define ERROR(x,...)error("error: " ,x,## __VA_ARGS__)

#define SMALLOC(x,y){x=(y*)malloc(sizeof(y));\
						if( x == NULL) ERROR("malloc out of memory: %s",strerror(errno));}





						
struct p0fsig{
	u_short win;
	u_char ttl;
	u_char df;
	u_char options[OPTBUFF_SIZE];
	u_int options_size;
};

struct pcap_parameter{
	int linktype;
	libnet_t *libnet_h;
	struct p0fsig p0f;
};


struct tcpopt{
	u_char *eool;
	u_char *noop;
	u_char *mss;
	u_short *sakp;
	u_char *wss;
	u_char *ts;
	
	short align;

};





struct ip_packet{ 
	u_int  headers_size;
	struct ip *ip;
	u_char *ip_options;
	u_int  ip_options_size;
	
	u_char *layer4;
	u_char *layer4_options;
	u_int  layer4_options_size;
	struct tcpopt decoded_tcp_options;
	u_char *payload; 
	u_int  payload_size;
	
	u_char raw[DEFAULT_SNAPLEN];
	
};




void* authenticate_syns(char *ifname,u_int32_t host,char **p0fstr,char *filter);
void pcap_callback(u_char *param, const struct pcap_pkthdr *header, const u_char *pkt_data);

int verbose;
int flag;






void error(char *fn,char *message, ...){
   
   va_list ap;
   char errmsg[1024];

  	va_start(ap, message);
  	vsnprintf(errmsg, sizeof(errmsg), message, ap);
  	va_end(ap);
   
	fprintf(stderr,"%s%s\n",fn,errmsg);
	exit(0);
}





u_int ip_to_int(char *ip,int *err){ 

	int a,
		c=0,
		pos=0,
		tmpint;
		
	char t[4];
	u_char  addr[4];
	
	for(a=0; pos<4; a++){	
		if(ip[a]=='.' || ip[a]==0){
		  	t[c]=0;
		  	tmpint=strtol(t, (char **)NULL, 10);
    		if(errno==EINVAL || (tmpint<0 || tmpint>255) )goto bad;
    		addr[pos]=(u_char)tmpint;
			pos++;
			c=0;
		}else {
			t[c]=ip[a];
			c++;
		}
	}
	
	if(pos!=4)goto bad;
	if(err!=NULL)*err=0;
	return *(u_int*)addr;
	
bad:
	if(err!=NULL)*err=1;
	return 0;
}





char *int_to_ip(u_int32_t ip){
	u_char *tmp=(u_char *)&ip;
	static char ret[16];
	sprintf(ret,"%d.%d.%d.%d",tmp[0] & 0xff,tmp[1] & 0xff,tmp[2] & 0xff,tmp[3] & 0xff);
	return ret;
	
}





 
int loadtcpoptions(struct ip_packet *pkt,char *buff,int buffsize){
  
  u_char kind;
  struct tcpopt *opts;
  int a;
  
  pkt->decoded_tcp_options.eool=NULL;
  pkt->decoded_tcp_options.noop=NULL;
  pkt->decoded_tcp_options.wss=NULL;
  pkt->decoded_tcp_options.mss=NULL;
  pkt->decoded_tcp_options.ts=NULL;

  for(a=0;a<buffsize;){
   kind=*(buff+a); 

	switch(kind){
		case TCPOPT_EOOL:
		  pkt->decoded_tcp_options.eool= (buff+a);
		  a++;
		  break;
		case TCPOPT_NOOP:
		  pkt->decoded_tcp_options.noop=(buff+a);
		  a++;
		  break;
		case TCPOPT_SACKP:
		  pkt->decoded_tcp_options.sakp=(u_short*)(buff+a);
		  a+=2;
		  break;
		case TCPOPT_MSS:
		  pkt->decoded_tcp_options.mss=(buff+a);
		  a+=4;
  		  break;
		case TCPOPT_WSS:
		  pkt->decoded_tcp_options.wss=(buff+a);
		  a+=3;
		  break;
		case TCPOPT_TS:
		  pkt->decoded_tcp_options.ts=(buff+a);		  
		  a+=10;
		  break;
		default:
			//printf("unknown tcp option %d\n",kind);
			a+=(*(buff+a+1));

	}
  }
   
  return 1;
  }
  
  
  
  
  
  




int load_ip_packet(const u_char* pkt_data,struct ip_packet* pkt,u_int rawlen,int linktype){
	int size_link=get_link_size(linktype);
	u_int l4h_off;
	u_int ip_hl;
	
	pkt->ip_options_size=0;
	pkt->layer4_options_size=0;
	

	memcpy(pkt->raw + BALIGN ,pkt_data,rawlen);
	
	pkt->ip = (struct ip *) (BALIGN + pkt->raw + size_link);
	
	//ip header len
	ip_hl=pkt->ip->ip_hl << 2; //no ntohl
	// decoding ip options
	if(ip_hl > LIBNET_IPV4_H){
		pkt->ip_options=(u_char*) (BALIGN + pkt->raw + size_link + LIBNET_IPV4_H);
		pkt->ip_options_size= ip_hl - LIBNET_IPV4_H;
	}


	//deve essere castato di volta in volta da ogni decoder di livello 4
	pkt->layer4 = (BALIGN + pkt->raw + size_link + LIBNET_IPV4_H + pkt->ip_options_size);
	
	pkt->payload_size=( ntohs(pkt->ip->ip_len) ) -  LIBNET_IPV4_H - pkt->ip_options_size;
	pkt->headers_size=LIBNET_IPV4_H + pkt->ip_options_size; 
	
	pkt->payload=pkt->layer4;
	
	
	  switch (pkt->ip->ip_p) {
	
		case IPPROTO_TCP: 
			pkt->payload += LIBNET_TCP_H;
			pkt->layer4_options =  pkt->payload;
			l4h_off = ( ((struct tcphdr*)pkt->layer4)->th_off) << 2; //no ntohl
			// decoding tcp options
			if(l4h_off > LIBNET_TCP_H){
			  pkt->layer4_options_size = l4h_off - LIBNET_TCP_H;
			  pkt->payload += pkt->layer4_options_size;
			  loadtcpoptions(pkt,pkt->layer4_options,pkt->layer4_options_size);
			}
			pkt->payload_size -= (LIBNET_TCP_H + pkt->layer4_options_size);
			pkt->headers_size += (LIBNET_TCP_H + pkt->layer4_options_size);
			((struct tcphdr*)pkt->layer4)->th_sum = 0;
			break;
			
 		case IPPROTO_UDP: 
 			//unused
 			break;
 			
 		case IPPROTO_ICMP:
	 		//unused
 			break;
 	  }
 	  
 	//mai far puntare pk->payload a NULL..infatti mi serve per sapere la lunghezza degli headers
 	
 	//handle (odd) malformed packets 
 	if( pkt->payload_size < 0) pkt->payload_size = 0;
 	
 	
 	return 1;
}






int build_ip(struct ip_packet *pk,libnet_t *libnet_hnd){
	libnet_ptag_t lnip;
	
	if(pk->payload_size>0)
		libnet_build_data(pk->payload, pk->payload_size, libnet_hnd, 0);	
		//libnet_build_payload(pk->payload, pk->payload_size, libnet_hnd, 0);//libnetNG
		
	if(pk->payload != pk->layer4)//ip packet with layer4
	  switch (pk->ip->ip_p) {
		case IPPROTO_TCP:  
 			build_tcp(pk,libnet_hnd);
 			break;
 		case IPPROTO_UDP:
 			//unused
  			
 		case IPPROTO_ICMP:
 			//unused
  			
 		default: return 0;
 		
 		
 	  }
 	
 	if(pk->ip_options_size>0)
 		libnet_build_ipv4_options(pk->ip_options,pk->ip_options_size,libnet_hnd,0);
 		
    lnip = libnet_build_ipv4(                                                                          
           pk->headers_size + pk->payload_size,  	     // length 
           pk->ip->ip_tos,                               		// TOS  
           ntohs(pk->ip->ip_id),                 						// IP ID 
           ntohs(pk->ip->ip_off),                                  	// IP Frag 
           pk->ip->ip_ttl,                                 		//TTL 
           pk->ip->ip_p,                        				// protocol 
           0,                                  					// checksum 
           pk->ip->ip_src.s_addr,       						// source IP 
           pk->ip->ip_dst.s_addr,       						// destination IP 
           NULL,                               					// payload 
           0,                                  					// payload size 
           libnet_hnd,                  						// libnet handle 
           0);													// libnet ptag
           
     
     if(lnip==-1)
     	ERROR("Eth forging error: %s", libnet_geterror(libnet_hnd));
     
     return 1;
}





int build_tcp(struct ip_packet *pk,libnet_t *libnet_hnd){
			
	libnet_ptag_t lntcp;
	u_int tcp_size= LIBNET_TCP_H + pk->payload_size  + pk->layer4_options_size;
	struct tcphdr* tcp=(struct tcphdr*) pk->layer4;
	

	if(pk->layer4_options_size>0)
		libnet_build_tcp_options((u_int8_t *)pk->layer4_options, pk->layer4_options_size, libnet_hnd, 0);
	
	
	lntcp = libnet_build_tcp(
        ntohs(tcp->th_sport),            // source port 
        ntohs(tcp->th_dport),            // destination port 
        ntohl(tcp->th_seq),              // sequence number 
        ntohl(tcp->th_ack),              // acknowledgement num 
        tcp->th_flags,                   // control flags 
        ntohs(tcp->th_win),                  	 // window size 
        ntohs(tcp->th_sum),  			                     // checksum 
        ntohs(tcp->th_urp),                     // urgent pointer 
        tcp_size,       // TCP packet size 
	    NULL,                    			 // payload 
        0,                       			 // payload size 
        libnet_hnd,
        0);

    if(lntcp==-1)
    	ERROR("tcp forging error: %s", libnet_geterror(libnet_hnd));


    return lntcp;
  }
  







char *nexttoken(char **str,char sep){
 #define CONTINUE_IF_QUOTED(x,y) if( (x>y) && (*(x-1)=='\\'))continue;
	int a;
	char *t=*str,*sw;
	//scorre str usando t come puntatore
	if(*t!=0)
	for(;;t++){

		CONTINUE_IF_QUOTED(t,*str);
		//se trovo un apice sposto il puntatore fino al prossimo apice saltando quelli quotati
		if(*t=='\'')
			for(t++;;t++){
				CONTINUE_IF_QUOTED(t,*str);
				//if( (t>*str) && (*(t-1)=='\\')) continue;
				if(*t == '\'')break;
			}
		//se trovo un sep lo sostituisco con zero,inverto str e t in modo da far puntare str 
		//al primo carattere dopo lo zero e t all'inizio della stringa		
		if(*t == sep){
			*t = 0;
			sw = *str;
			*str = ++t;
			return sw;
		}
		//se la stringa e' finita inverto str e t in modo da far puntare str a zero e t
		//all'inizio della stringa
		if(*t == 0){
			sw = *str;
			*str = t;
			return sw;
		}
	}
	return NULL;
}


	
	





int get_link_size(int dl){
	   switch (dl) {    
        case 	DLT_EN10MB:	return 14;
        case  	DLT_LOOP:	return 4;
        default:return -1;
        }
 }
 
 
 
 
 
 char *meminmem(char *str,u_int str_size,char *buff,u_int buff_size){
    char *off=NULL;    
    
    int a;
    
    if(buff_size == 0) return NULL;
    
    //find offset per la prima occorrenza di str
   	for(a=0;a<buff_size;a++)
		if(buff[a]==str[0])
			if(!memcmp(str,(buff+a),str_size)){
				off = ( buff + a);
				break;
			}
	
    return off;
 }



//sostituisce una str con un'altra un un buffer
int memreplace(char *str, u_int str_size, char* rplc, u_int rplc_size, char *buff, u_int buff_size,int limit){
    char *off=NULL;

    char *end= (buff+buff_size);
    int len,cnt=0;
    
    if(buff_size == 0) return 0;
    
    //find offset per la prima occorrenza di str

	off=meminmem(str,str_size,buff,buff_size);
    if(off == NULL)return 0;
    
    //fai spazio nel buffer per il replacement (che puo essere piu grande o piu piccolo di str)
    len = end - off;  
    memmove( off + rplc_size, off + str_size, len); 
    memcpy(off,rplc, rplc_size);
	//cnt++;
	//se il buffer non e' finito e il limite de replc imposto non e' raggiunto applica la ricorsione
	limit--;
    if(len+(rplc_size-str_size) >= str_size && limit!=0 )	
		cnt += memreplace(str, str_size, rplc, rplc_size, off + rplc_size,len+((rplc_size-str_size)) ,limit);		
	
	//ritorna il contatore tel tot dei rplc
	return ++cnt;
 }
 
 
 
 
 int adjust_tcp_options(u_char *buff,int buffsize,struct ip_packet *pkt){
  
  u_char kind;
  struct tcpopt *opts;
  int a=0,size;
  time_t tsval;
  u_short tshort;
 
  while(a < buffsize){
   kind=*(buff+a); 
	 switch(kind){
		case TCPOPT_EOOL:
		  size=1;
		  break;
		case TCPOPT_NOOP:
		  size=1;
		  break;
		case TCPOPT_SACKP:
		  size=2;
		  break;
		case TCPOPT_MSS:
		  size=4;
		  if(*(buff+a+1)==0){
		  	if (pkt->layer4_options_size > 0 || pkt->decoded_tcp_options.mss == NULL) {
		      tshort=htons(1460);
			  if(verbose)printf( "forcing MSS to:1460\n");
			  *(buff+a+1)=TCPOPT_MSS;
			  memcpy(buff+a+2,&tshort,2);	
			  break;
		    }
		  	memcpy( buff+a, pkt->decoded_tcp_options.mss,size);
		  }
  		  break;
		case TCPOPT_WSS:
		  size=3;
		  if(*(buff+a+1)==0){
		    if (pkt->layer4_options_size > 0 || pkt->decoded_tcp_options.wss == NULL) {
			  if(verbose)printf( "forcing WSS to:0\n");
			  *(buff+a+1)=TCPOPT_WSS;
			  *(buff+a+2)=0;
			  break;
		    }
		  	memcpy( buff+a, pkt->decoded_tcp_options.wss,size);
		  }
		  break;
		case TCPOPT_TS:
		 size=10;
		 if(*(buff+a+1)==0) {
		    if (pkt->layer4_options_size == 0 || pkt->decoded_tcp_options.ts == NULL) {
		      tsval=htonl(time(NULL));
			  if(verbose)printf( "forcing timestamp to:%d.0\n",ntohl(tsval));
			  *(buff+a+1)=TCPOPT_TS;
			  memset(buff+a+2,0,8);	
			  memcpy(buff+a+2,&tsval,4);
			  break;
		    }
		    memcpy( buff+a, pkt->decoded_tcp_options.ts,size);
	     }
		  break;
		  
		  default:
		    printf("ADJUST: STRANGE ERROR! unknown tcp option %d\n",kind);
			//exit(0);
			size=*(buff+a+1);
	}
	a+=size;
  }

   
  return 1;
  }
 

int a_buildtcpopts(char **str,u_char *res,u_int *optsize){

	u_char 	*ap;
	int a=0,
		b;
	u_int ival;
	u_short sval;
	char tmp[20];
	
		memset(res,0,40);

         while((ap = (char*)nexttoken((char**)str, ',')) != NULL){
         	if(!strcmp(ap,"N")){
       			*(res+a)=TCPOPT_NOOP;  		
       			a++;
       			continue;
         	}
          	if(!strcmp(ap,"E")){
       			*(res+a)=TCPOPT_EOOL;  		
       			a++;
       			continue;
         	}
         	if(!strcmp(ap,"sackp")){
       			*(res+a)=TCPOPT_SACKP;
       			*(res+a+1)=2;
       			a+=(*(res+a+1));
       			continue;
         	}
         	
         	if(!strncmp(ap,"W",1)){
       			*(res+a)=TCPOPT_WSS;
       			if(!strncmp( (ap+1),"*",1)){
       				*(res+a+1)=0;
       				a+=3;
       				continue;
       			}
       			*(res+a+1)=3;
       			*(res+a+2)=(u_char)strtol((ap+1), (char **)NULL, 10);
       			if(errno==EINVAL)
						ERROR("tcp.options.wss: wrong integer value");
       			a+=3;
       			continue;
         	}
         	if(!strncmp(ap,"M",1)){
       			*(res+a)=TCPOPT_MSS;       			
       			if(!strncmp( (ap+1),"*",1)){
       				*(res+a+1)=0;
       				a+=4;
       				continue;
       			}
       			*(res+a+1)=4;
       			sval=(u_short)strtoul((ap+1), (char **)NULL, 10);
       			if(errno==EINVAL)
						ERROR("tcp.options.mss: wrong integer value");
				sval=htons(sval);
				memcpy((res+a+2),&sval,2);
				a+=4;
       			continue;
         	}
           	if(!strncmp(ap,"T",1)){
   
       		    *(res+a)=TCPOPT_TS;       			
   
   				if(!strcmp( ap,"T")){
       				*(res+a+1)=0;
       				a+=10;
       				continue;
       			}
       			
       			*(res+a+1)=10;
       			for(b=0;  ;b++){
       				
					if( *(ap+1+b) == '.'){
					  *(tmp+b)=0;
					  ival=(u_int)strtoul(tmp, (char **)NULL, 10);
					  if(errno==EINVAL)
						ERROR("tcp.options.ts[0]: wrong integer value");
					  ival=htonl(ival);
					  memcpy((res+a+2),&ival,4);
					  ap+=b+1;
					  b=0;
					}
	
					if( *(ap+1+b) == 0 ){
					  *(tmp+b)=0;
					  ival=(u_int)strtoul((tmp), (char **)NULL, 10);
					  if(errno==EINVAL)
						ERROR("tcp.options.ts[1]: wrong integer value");
					  ival=htonl(ival);
					  memcpy((res+a+2+4),&ival,4);
					  break;
					 }
					
				*(tmp+b)=*(ap+1+b);
	
       			}	
       			a+=10;
       			continue;
       		
         	}

         	ERROR("unknown tcp option:%s",ap);
         }
*optsize=a;
return 1;
}

 

 
 
void usage(char *appname){
	printf(	"authsyn %s\n"
			"  usage:\n"
			"    %s -i <interface> -h <host> -s <p0f signature> [-f <pcapfilter>] [-v]\n\n"
			"      -i      output interface\n"
			"      -h      destination host\n"
			"      -s      signature from p0f databse\n"
			"      -f      pcap filter string\n"
			"      -v      be verbose\n\n"
			"    example: %s -i eth0 -h 85.41.32.67 -s \"16384:64:1:60:M*,S,T,N,W0:.:-Linux:2.2 (Opera?)\" -f \"dst port 80\"\n\n"
			"  note: if * is used, the original value is preserved\n"
			"      it's possible to force T value by specifing a pair of values (tsval and echo-reply) separated by a dot, ex T1139773350.0\n\n"
			,VERSION,appname,appname);
	exit(0);
}




int main(int argc, char **argv){
	
	int c,err;

	char *ifname=NULL,
		 *p0fsig=NULL,
		 *filter=NULL;	

    u_int32_t host=0;
    extern char *optarg;

	
	
	verbose = 0;
	
     while ((c = getopt(argc, argv, ":vi:h:s:f:")) != -1) {
       switch(c) {
         case 's':
         	p0fsig= optarg;
         	break;
         case 'f':
         	filter=optarg;
         	break;
         case 'v':
         	verbose=1;     
            break;
         case 'i':
         	ifname = optarg;
			break;
		 case 'h':
			host=ip_to_int(optarg,&err);
			if(err)ERROR("parsing error:%s",optarg);
			break;
         case ':':
            printf("error: -%c: argument is mandatory \n", optopt);
            usage(argv[0]);
            break;
         case '?':
            printf("error: unknown option %c\n", optopt);
            usage(argv[0]);
            break;
       }
     }
	
	if ( host == 0 || ifname == NULL || p0fsig == NULL ) usage(argv[0]);
	

	authenticate_syns(ifname, host,&p0fsig,filter);

	return 0;	
}



void* authenticate_syns(char *ifname,u_int32_t host,char **p0fstr,char *pcapfilter){
	//ioctl parameter
	int n=1;
	
	//pcap handle
	pcap_t *pcap_hnd;
	//pcap errbuff
 	char p_errbuf[PCAP_ERRBUF_SIZE];
 	
 	//pcapp filter
 	struct bpf_program filter;
	u_int32_t net = 0;        
    u_int32_t mask = 0;       
    
 	char filterstring[512];
 	
 	//libnet handle
	libnet_t *libnet_hnd;
	char ln_errbuf[LIBNET_ERRBUF_SIZE];

	//struct for reading l2addr with libnet
	struct libnet_ether_addr *hwaddr;
	//parameter for pcap callbak
	struct pcap_parameter pcp;
	
	struct p0fsig p0f;
	
	int token=0;
	char *ap;
	
	
    //init libnet
  libnet_hnd = libnet_init(LIBNET_RAW4,ifname,ln_errbuf);
  if(libnet_hnd==NULL)
	ERROR("libnet init failed: %s",ln_errbuf);
	

  sprintf(filterstring,"(tcp[tcpflags] & (tcp-syn) = tcp-syn and (dst host %s))",int_to_ip(host));
 	
  if(pcapfilter != NULL)
	 sprintf(filterstring,"%s and (%s)",filterstring,pcapfilter);

  if(verbose)printf("PCAP FILTER: %s\n",filterstring);
  
  pcap_hnd=pcap_open_live(ifname, DEFAULT_SNAPLEN, 1, 0, p_errbuf);
 
  if(pcap_hnd==NULL)
	ERROR("pcap error:%s",p_errbuf);
	
  
  #ifdef BSD
   ioctl(pcap_fileno(pcap_hnd), BIOCIMMEDIATE, &n);
  #endif

  if (pcap_lookupnet(ifname, &net, &mask, p_errbuf) < 0) 
       ERROR("error looking up network: %s", p_errbuf);

 

 //compile pcap filter
 if((pcap_compile(pcap_hnd, &filter, filterstring, 0,mask))==-1)
  	ERROR("pcap filter error: ",filterstring);
 //set the filter 	
 pcap_setfilter(pcap_hnd, &filter);
 
 
 //pcap parameter
 pcp.libnet_h=libnet_hnd;
 pcp.linktype=pcap_datalink(pcap_hnd);
 
 while((ap = (char*)nexttoken(p0fstr, ':')) != NULL){
   switch(token){
    case 0:
 	  p0f.win=strtoul((ap), (char **)NULL, 10);
      if(errno==EINVAL)ERROR("Window: wrong integer value");
 	  break;
 	case 1:
 	  p0f.ttl=strtoul((ap), (char **)NULL, 10);
 	  if(errno==EINVAL)ERROR("TTL: wrong integer value");
 	  break;
 	case 2:
 	  p0f.df=strtoul((ap), (char **)NULL, 10);
 	  if(errno==EINVAL)ERROR("DF: wrong integer value");
 	  if( p0f.df !=0 && p0f.df != 1)
 	  	ERROR("parser error: DF flag can be set to 0 or 1 only");
 	  break;
 	case 4:
 	  p0f.options_size=sizeof(p0f.options);
 	  a_buildtcpopts(&ap,p0f.options,&p0f.options_size);
 	  break;
 	}
 	token++;

 }
 pcp.p0f=p0f;
 
 printf("ready\n");
 
 //use a flag to avoid pcap to see the new generated packet
 flag =0;
 pcap_loop(pcap_hnd, 0, pcap_callback, (u_char*)&pcp);


 pcap_close(pcap_hnd);
  
  
}


 
void pcap_callback(u_char *param, const struct pcap_pkthdr *header, const u_char *pkt_data){
	//casted parameter for this function
	struct pcap_parameter *pcp=(struct pcap_parameter *) param;

	//new ip packet
	struct ip_packet *newpkt;
	
	u_char tmp_layer4_options[OPTBUFF_SIZE];


    if ( flag == 1 ){
    	flag =0;
    	return;
    }

	SMALLOC(newpkt,struct ip_packet);
	if( (load_ip_packet(pkt_data,newpkt,header->caplen,pcp->linktype)) == 0 ){ 
    	free(newpkt);
    	printf("packet ignored\n");
		return;
	}
	
	
	if(pcp->p0f.df == 1)newpkt->ip->ip_off |= (IP_DF & IP_OFFMASK) ;
	if(pcp->p0f.df == 0)newpkt->ip->ip_off &= (IP_DF & IP_OFFMASK);

	((struct tcphdr*)newpkt->layer4)->th_win = htons(pcp->p0f.win);
	newpkt->ip->ip_ttl=pcp->p0f.ttl;

	memcpy(tmp_layer4_options,pcp->p0f.options,pcp->p0f.options_size);
	adjust_tcp_options(tmp_layer4_options,pcp->p0f.options_size,newpkt);
	
	if( newpkt->layer4_options_size != pcp->p0f.options_size ){
		//shift payload after the end of new options
		memmove( newpkt->layer4_options + pcp->p0f.options_size, newpkt->payload, newpkt->payload_size); 
		newpkt->payload += ( pcp->p0f.options_size - newpkt->layer4_options_size );
		newpkt->headers_size += ( pcp->p0f.options_size - newpkt->layer4_options_size );
		newpkt->layer4_options_size=pcp->p0f.options_size;
	}

	memcpy(newpkt->layer4_options,tmp_layer4_options,pcp->p0f.options_size);
	

 			
	if(!build_ip(newpkt,pcp->libnet_h))
		ERROR("UNEXPECTED ERROR 2"); 
	
	if(libnet_write(pcp->libnet_h)==-1)
		ERROR("Write error: %s", libnet_geterror(pcp->libnet_h));
		

	libnet_clear_packet(pcp->libnet_h);
	flag=1;
	
	free(newpkt);
	
	if(verbose)printf("SYN packet authenticated\n");
				
}

